/* ------------------------------------------------------------------------- *
 * ArdIO - プログラマブルタイマー基板 ベースファームウェア  v1.00 2019-05-14 *
 * Copyright (C) 2018-2019 KYOHRITSU ELECTRONIC INDUSTRY CO., LTD.           *
 *     http://prod.kyohritsu.com/KP-TMAR08.html                              *
 * 本プログラムは共立電子産業「プログラマブルタイマー基板」(KP-TMAR08) での  *
 * 利用に限り、自由に使用・改変することができます。                          *
 * プログラムの内容に関する質問・サポート等は行っておりません。              *
 * ------------------------------------------------------------------------- */
/* --------------------------------------------------------------------------*
 * 本ソフトウェアは下記のオープンソースソフトウェアを利用しています。        *
 *   - TimerOne                                                              *
 *     by Jesse Tane, Jérôme Despatis, Michael Polli, Dan Clemens,           *
 *     and Paul Stoffregen, under the terms of CC BY 3.0 US license          *
 *     http://creativecommons.org/licenses/by/3.0/us/                        *
 * ------------------------------------------------------------------------- */
#include <limits.h>
#include <avr/wdt.h>
#include <TimerOne.h>
#include "Common.h"
#include "RingBuffer.h"
#include "Parser.h"
#include "WireCode.h"
#include "TimerCore.h"

RingBuffer rxb(RXB_LENGTH);
Parser ps(LINEB_LENGTH);
Token t0, t1;
WireCode wc(WIRECODE_BLOCK_SIZE, NUM_BLOCKS);
TimerCore timer;
volatile bool timerTriggered = false;
volatile bool running = false;
volatile union { int n; char b[2]; } wdtResetCount;

// ウォッチドッグリセット継続カウンターをリロード
inline void reloadWdtResetCount() {
  // 上位バイトへ直接代入
  wdtResetCount.b[1] = WDT_RESET_TIMEOUT_X256MS;
}

// 起動メッセージを表示
void printStartupMessages() {
  Serial.println(F("ArdIO Macro Processor - Version " ARDIO_VERSION_STRING));
  Serial.println(F("(C) 2018-2019 KYOHRITSU ELECTRONIC INDUSTRY CO., LTD."));
}

// タイマー実行停止
void stop() {
  running = false;

  // リレー出力をオフ
  for (int i = 0; i < 4; i++) {
    digitalWrite(PIN_RELAY[i], 0);
  }
  // ブロックを未指定状態に戻す
  wc.setBlockId(-1);
  wc.resetErrorFlags();
  Serial.println(F("Stop"));
}

// タイマー実行開始
void run(int _mode) {
  int mode = _mode & 0x03;

  // ブロックを未指定状態に戻す & 最後の未保存分を保存
  if (wc.getBlockId() != -1) {
    wc.saveBlock();
    wc.setBlockId(-1);
  }
  Serial.print(F("Run <Mode "));
  Serial.print(mode);
  Serial.println(F(">"));
  wc.loadBlock(mode);
  timer.clear();
  timer.loadWireCode(wc.getBlob());
  timer.reset();
  running = true;
}

// モード選択ジャンパー (M1, M0) を読み取り、現在のモード選択状態を取得
int getMode() {
  return (digitalRead(PIN_JUMPER[5]) ? 0b10 : 0b00)
         | (digitalRead(PIN_JUMPER[4]) ? 0b01 : 0b00);
}

void setup() {
  // ウォッチドッグ有効化
  wdt_enable(WDTO_250MS);
  reloadWdtResetCount();
  // 入出力ピンのセットアップ
  for (int i = 0; i < 4; i++) {
    pinMode(PIN_RELAY[i], OUTPUT);
    digitalWrite(PIN_RELAY[i], 0);
    pinMode(PIN_PHOTOCOUPLER[i], INPUT_PULLUP);
  }
  for (int i = 0; i < 6; i++) {
    pinMode(PIN_JUMPER[i], INPUT_PULLUP);
  }
  // シリアルポート使用開始
  Serial.begin(SERIAL_BAUD);
  // タイマー初期化、割り込み有効化
  Timer1.initialize(1000);
  Timer1.attachInterrupt(timerOneCallback);
  Timer1.start();
  // 起動メッセージ表示
  printStartupMessages();
  // 現在のモード選択ジャンパーに応じたモードを実行開始
  run(getMode());
}

void loop() {
  char c;
  static int lineCount = 0;

  // ウォッチドッグ維持: カウンター値をリロード
  reloadWdtResetCount();
  // 受信バッファから構文解析バッファへ文字を転送
  if (rxb.available()) {
    rxb.get(c);
    if (ps.put(c)) {
      lineCount++;
      ps.setLineCount(lineCount);
      wc.setLineCount(lineCount);
    }
  }
  if (running) {
    // シリアル受信から STOP の入力があれば、実行停止
    if (ps.checkStop()) {
      stop();
    }
    if (timerTriggered) {
      timerTriggered = false;
      timer.step();
    }
  } else {
    // シリアル受信から RUN の入力があれば、モード選択に応じて実行再開
    if (ps.checkRun()) {
      lineCount = 0;
      ps.setLineCount(lineCount);
      wc.setLineCount(lineCount);
      run(getMode());
    } else {
      if (ps.parse(t0, t1)) {
        wc.push(t0, t1);
      }
    }
  }
}

// TimerOne割り込みコールバック関数 (呼び出し間隔: 1ms)
void timerOneCallback() {
  static int tickCount = TIMER_TICK_MS;
  static bool acceptRx = true;
  char c;

  // ウォッチドッグリセット処理
  if (wdtResetCount.n > 0) {
    // カウンター値が残っていれば、リセットを実行
    wdtResetCount.n--;
    wdt_reset();
  } else {
    // カウンターが切れている場合は、永久ループに移行し
    // ウォッチドッグによる初期化を待機
    while (true);
  }
  // シリアル受信データが有れば、受信バッファに1文字格納する
  if (Serial.available()) {
    c = Serial.read();
    // セミコロン (コメント開始) の場合、格納を中断
    if (c == ';') {
      acceptRx = false;
    }
    // 改行文字 (行末) の場合、格納を再開
    if (c == '\r' || c == '\n') {
      acceptRx = true;
    }
    if (acceptRx) rxb.put(c);
  }
  // タイマー実行時間のカウント
  if (--tickCount == 0) {
    timerTriggered = true;
    tickCount = TIMER_TICK_MS;
  }
}

