#include <Arduino.h>
#include <EEPROM.h>
#include <string.h>
#include "WireCode.h"

// 即値テーブル領域のアドレス原点
int WireCode::INTEGER_IMM_ADDRESS_BASE = 0x00;

// 即値テーブルサイズ: アドレス範囲 0x00..0x3f の2バイト32組
int WireCode::NUM_IMM_SLOTS = 32;

WireCode::WireCode(int _blockSize, int _numBlocks) {
  if (_blockSize > 0) {
    this->blockSize = _blockSize;
    this->blob = new unsigned char[_blockSize];
  } else {
    this->blob = nullptr;
  }
  this->numBlocks = _numBlocks > 0 ? _numBlocks : 1;
  this->blockId = -1;
  this->lineCount = 0;
  this->immediateCount = 0;
  this->resetErrorFlags();
}

WireCode::~WireCode() {
  delete[] this->blob;
}

void WireCode::clearBlob() {
  if (this->blob) {
    memset(this->blob,
           (int) TokenType::O_NOT_CONNECTED & 0xff,
           this->blockSize);
    this->immediateCount = 0;
  }
}

void WireCode::resetErrorFlags() {
  this->errorFlagWOutOfBlock = false;
}

void WireCode::setLineCount(int _n) {
  this->lineCount = _n;
}

int WireCode::getBlockId() {
  return this->blockId;
}

void WireCode::setBlockId(int _blockId) {
  this->blockId = _blockId;
}

int WireCode::findImmediateIndex(int _x) {
  int idx = -1;
  int k;
  unsigned char bLow, bHigh;

  if (_x == 0 || _x == 1) {
    // 0 と 1 は即値テーブル記録対象外なので検索失敗を返す
    idx = -1;
  } else {
#ifdef DEBUG
    Serial.print(F("ImmCount="));
    Serial.println(this->immediateCount);
#endif
    for (int i = 0; i < this->immediateCount
         && i < WireCode::NUM_IMM_SLOTS; i++) {
      bLow = this->blob[INTEGER_IMM_ADDRESS_BASE + i * 2];
      bHigh = this->blob[INTEGER_IMM_ADDRESS_BASE + i * 2 + 1];
      k = ((int) bHigh << 8) + bLow;
#ifdef DEBUG
      Serial.print(F("b["));
      Serial.print(INTEGER_IMM_ADDRESS_BASE + i * 2, HEX);
      Serial.print(F(", "));
      Serial.print(INTEGER_IMM_ADDRESS_BASE + i * 2 + 1, HEX);
      Serial.print(F("]="));
      Serial.println(k);
#endif
      if (_x == k) {
        idx = i;
        break;
      }
    }
  }
  return idx;
}

bool WireCode::getIndexedImmediate(int& _n, unsigned char _b) {
  bool valid = true;
  unsigned char bLow, bHigh;

  if (_b < WireCode::NUM_IMM_SLOTS + 2) {
    if (_b == 0 || _b == 1) {
      _n = _b;
    } else {
      bLow = blob[INTEGER_IMM_ADDRESS_BASE + (_b - 2) * 2];
      bHigh = blob[INTEGER_IMM_ADDRESS_BASE + (_b - 2) * 2 + 1];
      _n = ((int) bHigh << 8) + bLow;
    }
  } else if (_b != 0xff) {
    valid = false;
  }
  return valid;
}

bool WireCode::push(Token& _t0, Token& _t1) {
  int storeAddress;
  int storeValue;
  int idx;
  unsigned char bLow, bHigh;
  bool valid = true;

  if (_t0.isCommand()) {
    // ■コマンド
    if (_t0.type == TokenType::CMD_DUMP) {
      // ブロックの生データ出力
      this->dump();
      this->printVersion();
      Serial.println();
    } else if (_t0.getBlockId(this->numBlocks) != -1) {
      // ブロック開始
      if (valid) {
        this->setBlockId(_t0.getBlockId(this->numBlocks));
        this->clearBlob();
        this->errorFlagWOutOfBlock = false;
        Serial.print(F("Block "));
        Serial.print(this->blockId);
        Serial.print(F("> "));
      }
    } else if (_t0.type == TokenType::CMD_ENDBLOCK) {
      // ブロック終了: 現在のブロックを保存
      this->saveBlock();
      this->setBlockId(-1);
    }
  } else if (this->blockId == -1) {
    if (!_t0.isEmpty()) {
      valid = false;
      if (!valid) {
        if (!this->errorFlagWOutOfBlock) {
          // ブロック未指定
          Serial.print(F("Error[WB01] at line "));
          Serial.println(this->lineCount);
          Serial.println(F("Statement is out of block (first line here)"));
          this->errorFlagWOutOfBlock = true;
        }
      }
    }
  } else if (_t0.isBinaryInput() || _t0.isIntegerInput()) {
    // ■各種入力デバイス (バイナリ/整数)
    if (_t1.isImmediate()) {
      // ○即値の代入
      if (valid) {
        valid = _t1.numErr == NumericError::NO_ERROR;
      }
      if (valid) {
        if (_t1.value == 0 || _t1.value == 1) {
          // 0, 1 いずれかの即値
          storeValue = _t1.value;
        } else {
          // それ以外の即値
          if ((idx = this->findImmediateIndex(_t1.value)) != -1) {
            // 登録済みの即値であった場合、そのアドレスを保存する
#ifdef DEBUG
            Serial.print(F("*Imm. found at idx "));
            Serial.println(idx);
#endif
            storeValue = idx + 2;
          } else {
            // 未登録の場合、即値テーブルの空きを調べて新規登録
            if (valid) {
              valid = this->immediateCount < WireCode::NUM_IMM_SLOTS;
              if (!valid) {
                // 即値領域が満杯
                Serial.print(F("Error[WN01] at line "));
                Serial.println(this->lineCount);
                Serial.println(F("Number table is full"));
              }
            }
            if (valid) {
              // 即値を追加
              bLow = _t1.value & 0xff;
              bHigh = _t1.value >> 8;
#ifdef DEBUG
              Serial.print(F("*Imm. registered at idx "));
              Serial.println(this->immediateCount);
              Serial.print(F("*Imm.Store: b[0x"));
              Serial.print(WireCode::INTEGER_IMM_ADDRESS_BASE + this->immediateCount * 2, HEX);
              Serial.print(F(", 0x"));
              Serial.print(WireCode::INTEGER_IMM_ADDRESS_BASE + this->immediateCount * 2 + 1, HEX);
              Serial.print(F("]="));
              Serial.print(bLow);
              Serial.print(F(", "));
              Serial.println(bHigh);
#endif
              this->blob[WireCode::INTEGER_IMM_ADDRESS_BASE
                         + this->immediateCount * 2] = bLow;
              this->blob[WireCode::INTEGER_IMM_ADDRESS_BASE
                         + this->immediateCount * 2 + 1] = bHigh;
              storeValue = this->immediateCount + 2;
              this->immediateCount++;
            }
          }
        }
      }
      if (valid) {
        storeAddress = (int) _t0.type & 0x0ff;
        this->blob[storeAddress] = storeValue;
      }
    } else if (_t1.isBinaryOutput() || _t1.isIntegerOutput()) {
      // ○出力の代入
      if (valid) {
        storeAddress = (int) _t0.type & 0x0ff;
        storeValue = (int) _t1.type & 0x0ff;
        this->blob[storeAddress] = storeValue;
      }
    } else if (_t1.isBinaryInput() || _t1.isIntegerInput() || _t1.isNumericInput()) {
      // ×入力の代入
      if (valid) {
        valid = false;
        if (!valid) {
          // 入力を代入した
          Serial.print(F("Error[WT01] at line "));
          Serial.println(this->lineCount);
          Serial.println(F("INPUT devices cannot be declared as a source"));
        }
      }
    }
  } else if (_t0.isNumericInput()) {
    // ■数値限定の入力デバイス (バージョン表記の2個が該当)
    if (_t1.isImmediate()) {
      // ○即値の代入
      if (valid) {
        valid = _t1.numErr == NumericError::NO_ERROR;
      }
      if (valid) {
        if (_t1.value == 0 || _t1.value == 1) {
          // 0, 1 いずれかの即値
          storeValue = _t1.value;
        } else {
          // それ以外の即値
          if ((idx = this->findImmediateIndex(_t1.value)) != -1) {
            // 登録済みの即値であった場合、そのアドレスを保存する
#ifdef DEBUG
            Serial.print(F("*Imm. found at idx "));
            Serial.println(idx);
#endif
            storeValue = idx + 2;
          } else {
            // 未登録の場合、即値テーブルの空きを調べて新規登録
            if (valid) {
              valid = this->immediateCount < WireCode::NUM_IMM_SLOTS;
              if (!valid) {
                // 即値領域が満杯
                Serial.print(F("Error[WN01] at line "));
                Serial.println(this->lineCount);
                Serial.println(F("Number table is full"));
              }
            }
            if (valid) {
              // 即値を追加
              bLow = _t1.value & 0xff;
              bHigh = _t1.value >> 8;
#ifdef DEBUG
              Serial.print(F("*Imm. registered at idx "));
              Serial.println(this->immediateCount);
              Serial.print(F("*Imm.Store: b[0x"));
              Serial.print(WireCode::INTEGER_IMM_ADDRESS_BASE + this->immediateCount * 2, HEX);
              Serial.print(F(", 0x"));
              Serial.print(WireCode::INTEGER_IMM_ADDRESS_BASE + this->immediateCount * 2 + 1, HEX);
              Serial.print(F("]="));
              Serial.print(bLow);
              Serial.print(F(", "));
              Serial.println(bHigh);
#endif
              this->blob[WireCode::INTEGER_IMM_ADDRESS_BASE
                         + this->immediateCount * 2] = bLow;
              this->blob[WireCode::INTEGER_IMM_ADDRESS_BASE
                         + this->immediateCount * 2 + 1] = bHigh;
              storeValue = this->immediateCount + 2;
              this->immediateCount++;
            }
          }
        }
      }
      if (valid) {
        storeAddress = (int) _t0.type & 0x0ff;
        this->blob[storeAddress] = storeValue;
      }
    } else {
      // ×即値以外の代入
      valid = false;
      if (!valid) {
        // バージョンに即値以外を代入した
        Serial.print(F("Error[WV01] at line "));
        Serial.println(this->lineCount);
        Serial.println(F("Only numbers can be accepted for the VERSION section"));
      }
    }
  } else if (_t0.isBinaryOutput() || _t0.isIntegerOutput()) {
    // ■×各種出力デバイス (バイナリ/整数)
    valid = false;
    if (!valid) {
      // 出力に代入しようとした
      Serial.print(F("Error[WT02] at line "));
      Serial.println(this->lineCount);
      Serial.println(F("OUTPUT devices cannot be declared as a destination"));
    }
  }
  return valid;
}

void WireCode::saveBlock(int _blockId) {
  bool valid = true;

  if (valid) {
    valid = _blockId >= 0 && _blockId < this->numBlocks;
  }
  if (valid) {
    valid = (_blockId + 1) * this->blockSize <= (int) EEPROM.length();
  }
  if (valid) {
    Serial.print(F(" Update"));
    for (int i = 0; i < this->blockSize; i++) {
      EEPROM.update(_blockId * this->blockSize + i, this->blob[i]);
      if ((i & 0x0f) == 0x0f) {
        Serial.print('.');
      }
    }
    Serial.print(F(" OK. "));
    this->printVersion();
    Serial.println();
  }
}

void WireCode::saveBlock() {
  this->saveBlock(this->blockId);
}

void WireCode::loadBlock(int _blockId) {
  bool valid = true;

  if (valid) {
    valid = _blockId >= 0 && _blockId < this->numBlocks;
  }
  if (valid) {
    valid = (_blockId + 1) * this->blockSize <= (int) EEPROM.length();
  }
  if (valid) {
    for (int i = 0; i < this->blockSize; i++) {
      this->blob[i] = EEPROM.read(_blockId * this->blockSize + i);
    }
  }
}

void WireCode::dump() {
  Serial.print(F("     |"));
  for (int i = 0; i < 16; i++) {
    Serial.print(F(" +"));
    Serial.print(i, HEX);
  }
  Serial.println();
  Serial.println(F(" ----+-------------------------------------------------"));
  for (int i = 0; i < this->blockSize; i++) {
    if ((i & 0x0f) == 0x00) {
      Serial.print(i < 0x0010 ? F("  0") : F("  "));
      Serial.print(i, HEX);
      Serial.print(F(" | "));
    }
    if (this->blob[i] < 0x10) {
      Serial.print(F("0"));
    }
    Serial.print(this->blob[i], HEX);
    Serial.print(F(" "));
    if ((i & 0x0f) == 0x0f || i == this->blockSize - 1) {
      Serial.println();
    }
  }
}

void WireCode::printVersion() {
  unsigned char majB = blob[(int) TokenType::NI_MAJOR_VERSION];
  unsigned char minB = blob[(int) TokenType::NI_MINOR_VERSION];
  bool valid = true;
  int majV, minV;

  valid = this->getIndexedImmediate(majV, majB)
          && this->getIndexedImmediate(minV, minB);
  Serial.print(F("Ver. "));
  if (majB == 0xff && minB == 0xff) {
    Serial.print(F("n/a"));
  } else if (!valid) {
    Serial.print(F("err"));
  } else {
    // メジャーバージョン
    if (majB != 0xff) {
      Serial.print(majV);
    }
    // マイナーバージョン
    if (minB != 0xff) {
      Serial.print('.');
      Serial.print(minV);
    }
  }
}

unsigned char* WireCode::getBlob() {
  return this->blob;
}

